unit RRFGrid;
  (* Component to add my preferred features to tDBGrid.  The current such
    features are
        1.  Boolean entries are displayed as checkboxes.
        2.  The selected row is distinctly colored.
        3.  The user can click a title to select an index to re-sort the
               display.  Column titles that are associated with re-sorting
               are distinctively colored.

    New Properties:

           name             type     default          description
       NonSortTitleColor   tColor   clBtnFace   color of column titles that are
                                                  NOT associated with indices
       SortTitleColor      tColor    clLime     color of column titles that ARE
                                                  associated with indices
       SortOptions         tStrings   empty     each string names a field and
                                                  an index, separated by a
                                                  vertical bar.  A null index
                                                  name denotes the default
                                                  ordering.

    Notes:

       To get the title-coloring to work, the grid must have explicit fields,
      obtained via the Column Editor.
        A fatal exception is raised whenever one of these grids is created and
      the datasource does not exist.  To avoid this problem,
      be sure that either
              (1) the datasource is on the same form, or
              (2a) at execution time, the data module containing the data
                    source is loaded first by the application (*.DPR), and
              (2b) at design time, the data module is opened first.
     *)

  interface
    uses
      Classes  (* ComponentState, csDesigning, tShiftState, tStringList *),
      Controls,
      DBGrids  (* dgEditing, tColumn, tCustomDBGrid, tDBGrid, tDBGridOptions *),
      Forms    (* Application *),
      Graphics (* fsBold, tColor *),
      Grids    (* tGridDrawState *),
      Windows  (* DFC_BUTTON, DFCS_BUTTONCHECK, DFCS_CHECKED, DrawFrameControl,
                    tRect, VK_SPACE *);
                    (* eject *)
    type
      tRRFDBGrid = class(tDBGrid)
        private
          fNonSortTitleColor (* usual title color            *),
          fSortTitleColor    (* title color if sort key      *): tColor;
          SavedGridOptions   (* keep from editing checkboxes *): tDBGridOptions;
          SortFieldNames     (* parsed from fSortOptions     *),
          SortIndexNames     (* parsed from fSortOptions     *),
          fSortOptions       (* <fieldname> | <indexname>    *): tStrings;
          procedure ParseOptions;
          procedure SetColumnTitleStyle;
          procedure SetOptions(const Value: tStrings);
          procedure ToggleBoolean;
        protected
          procedure Loaded;                                     override;
        public
          procedure CellClick(Column: TColumn);                 override;
          procedure ColEnter;                                   override;
          procedure ColExit;                                    override;
          constructor Create(AnOwner: tComponent);              override;
          destructor Destroy;                                   override;
          procedure DrawColumnCell(const Rect: TRect;
                                   DataCol: Integer;
                                   Column: TColumn;
                                   State: TGridDrawState);      override;
          procedure KeyDown(var Key: Word; Shift: TShiftState); override;
          procedure SetUpColumns;
          procedure TitleClick(Column: TColumn);                override;
        published
          property NonSortTitleColor: tColor read  fNonSortTitleColor
                                             write fNonSortTitleColor
                                             default clBtnFace;
          property SortOptions: tStrings  read  fSortOptions
                                             write SetOptions;
          property SortTitleColor: tColor read  fSortTitleColor
                                          write fSortTitleColor
                                          default clLime;
          property Anchors;
          property Columns;
          property DataSource;
          property DefaultDrawing;
          property Enabled;
          property Font;
          property OnDblClick;
          property OnTitleClick;
          property Options;
          property PopupMenu;
          property ReadOnly;
          property ShowHint;
          property TabOrder;
          property TabStop;
          property TitleFont;
          property Visible;
      end (* tRRFDBGrid *);

  implementation
    uses
      DB         (* tFieldType *),
      DBTables   (* tTable *),
      Dialogs    (* ShowMessage *),
      MyConsts   (* nullstring, VBAR *),
      MyUtils    (* RunningInIDE *),
      SysUtils   (* trim *);

    procedure tRRFDBGrid.CellClick(Column: TColumn);
      (* On click, toggle the boolean *)

      begin (* tRRFDBGrid.CellClick *)
        if SelectedField.DataType = ftBoolean then
          ToggleBoolean
      end   (* tRRFDBGrid.CellClick *);

    procedure tRRFDBGrid.ColEnter;
      (* Save options and suppress editing *)

      begin (* tRRFDBGrid.ColEnter *)
        if SelectedField.DataType = ftBoolean then
          begin (* save options *)
            SavedGridOptions := Options;
            Options := Options - [dgEditing]
          end   (* save options *)
      end   (* tRRFDBGrid.ColEnter *);

    procedure tRRFDBGrid.ColExit;
      (* Restore options *)

      begin (* tRRFDBGrid.ColEnter *)
        if SelectedField.DataType = ftBoolean then
          Options := SavedGridOptions
      end   (* tRRFDBGrid.ColEnter *);

    constructor tRRFDBGrid.Create(AnOwner: tComponent);
      (* initialize properties *)

      begin (* tRRFDBGrid.Create *)
        inherited Create(AnOwner);
        fNonSortTitleColor := clBtnFace;
        fSortOptions := tStringList.Create;
        fSortTitleColor := clLime;
        SortFieldNames := tStringList.Create;
        SortIndexNames := tStringList.Create;
        TitleFont.Style := [fsBold]
      end   (* tRRFDBGrid.Create *);
      (* eject *)
    destructor tRRFDBGrid.Destroy;
      (* Release memory *)

      begin (* tRRFDBGrid.Destroy *)
        fSortOptions.Free;
        SortFieldNames.Free;
        SortIndexNames.Free;
        inherited Destroy
      end   (* tRRFDBGrid.Destroy *);

    procedure tRRFDBGrid.DrawColumnCell(const Rect: TRect;
                                        DataCol: Integer;
                                        Column: tColumn;
                                        State: tGridDrawState);
      (*   Cracker technique is described in FAQ711D on the Inprise website
         www.inprise.com/devsupport/delphi/qanda/FAQ711D.html.  It would be
         needed if similar code were being used in a noncomponent.
           Checkbox tricks from
         members.truepath.com/delphi/tips/tip30_tdbgridbooleancheckbox1.htm
      *)

      const
        CheckBoxes: array[boolean] of integer =
          (DFCS_BUTTONCHECK, DFCS_BUTTONCHECK or DFCS_CHECKED);

      var
        OldBrushColor, OldPenColor: tColor;

      begin (* tRRFDBGrid.DrawColumnCell *)
        if not RunningInIDE and (Column.Field.DataType = ftBoolean) then
          begin (* allow checkbox *)
            Canvas.FillRect(Rect);
            DrawFrameControl(Canvas.Handle,
                             Rect,
                             DFC_BUTTON,
                             CheckBoxes[Column.Field.AsBoolean])
          end   (* allow checkbox *)
        else
          with {tDBGridCracker(Sender),} Canvas do
            begin (* with *)
              OldBrushColor := Brush.Color;
              OldPenColor := Pen.Color;
              if DataLink.ActiveRecord = Row - 1 then
                if State = [] then
                  begin (* selected *)
                    Brush.Color := OldPenColor;
                    Pen.Color := OldBrushColor
                  end   (* selected *);
              DefaultDrawColumnCell(Rect, DataCol, Column, State);
              Brush.Color := OldBrushColor;
              Pen.Color := OldPenColor
            end   (* with *)
      end   (* tRRFDBGrid.DrawColumnCell *);

    procedure tRRFDBGrid.KeyDown(var Key: Word;
                                 Shift: tShiftState);
      (* Toggle boolean on space key *)

      begin (* tRRFDBGrid.KeyDown *)
        case Key of
          VK_SPACE: if SelectedField.DataType = ftBoolean then
                      ToggleBoolean
                    else
                      inherited KeyDown(Key, Shift)
               else inherited KeyDown(Key, Shift)
          end (* case *)
      end   (* tRRFDBGrid.KeyDown *);
      (* eject *)
    procedure tRRFDBGrid.Loaded;
      (* Utilize the properties loaded by Delphi *)

      begin (* tRRFDBGrid.Loaded *)
        inherited Loaded;
        SetUpColumns
      end   (* tRRFDBGrid.Loaded *);
      (* eject *)
    procedure tRRFDBGrid.ParseOptions;
      (* Generate SortFieldNames & SortIndexNames from SortOptions *)

      var
        I, VbarPos: integer;
        TheEntry, TheFieldName, TheIndexName: string;

      function EntryQualifies: boolean;
        (* Return true iff TheFieldName is a field name and TheIndexName is an
        index name.  This procedure is the site of an access-violation crash
        whenever a program using it is run from the IDE.  The crash comes when
        the program is starting to run, so protecting the try/except section
        within
            if not (csDesigning in ComponentState) then
        doesn't help.  The System.DebugHook code seems to be protective.  On
        one of the Delphi newsgroups, there was a rumor that DebugHook is
        NZ only when running inside the IDE *)

        var
          I: integer;

        begin (* EntryQualifies *)
          Result := true;
          if not RunningInIDE then
            try    (* check entry against DataSet *)
              with DataSource.DataSet as tTable do
                if Fields.FindField(TheFieldName) = nil then
                  Result := false
                else if TheIndexName <> nullstring then
                  begin (* must find index name *)
                    IndexDefs.Update;
                    for I := 1 to IndexDefs.Count do
                      if IndexDefs.Items[I-1].Name = TheIndexName then
                        Exit;
                    Result := false
                  end   (* must find index name *)
            except (* check entry against DataSet *)
              (* ignore *)
            end    (* check entry against DataSet *)
        end   (* EntryQualifies *);

      begin (* ParseOptions *)
        SortFieldNames.Clear;
        SortIndexNames.Clear;
        with fSortOptions do
          for I := 1 to Count do
            begin (* process one entry *)
              TheEntry := trim(Strings[I-1]);
              VbarPos := pos(VBAR, TheEntry);
              if VbarPos > 0 then
                begin (* line is <fieldname> | <indexname> *)
                  TheFieldName := trim(copy(TheEntry, 1, VbarPos-1));
                  TheIndexName := trim(copy(TheEntry, VBarPos+1, maxint))
                end   (* line is <fieldname> | <indexname> *)
              else
                begin (* just a field name *)
                  TheFieldName := TheEntry;
                  TheIndexName := nullstring
                end   (* just a field name *);

              if EntryQualifies then
                begin (* got one *)
                  SortFieldNames.Add(TheFieldName);
                  SortIndexNames.Add(TheIndexName)
                end   (* got one *)
            end   (* process one entry *)
      end   (* ParseOptions *);
      (* eject *)
    procedure tRRFDBGrid.SetColumnTitleStyle;
      (* Every column participating in sort options should be distinctive *)

      var
        I: integer;

      begin (* tRRFDBGrid.SetColumnTitleStyle *)
        for I := 1 to Columns.Count do
          with Columns[I-1] do
            if SortFieldNames.IndexOf(FieldName) <> -1 then
              Title.Color := fSortTitleColor
            else
              Title.Color := fNonSortTitleColor
      end   (* tRRFDBGrid.SetColumnTitleStyle *);

    procedure tRRFDBGrid.SetOptions(const Value: tStrings);
      (* Install the Value as SortOptions, and then parse SortOptions into
       field names and index names.  At the end, set the column headings to
       reflect sorting options *)

      begin (* tRRFDBGrid.SetOptions *)
{        if fSortOptions.Text <> Value.Text then}
          begin (* nondegenerate *)
            fSortOptions.BeginUpdate;
            try     (* move edited list in to active lists *)
              fSortOptions.Assign(Value)
            finally (* move edited list in to active lists *)
              fSortOptions.EndUpdate
            end     (* move edited list in to active lists *);

            SetUpColumns
          end   (* nondegenerate *)
      end   (* tRRFDBGrid.SetOptions *);
      (* eject *)
    procedure tRRFDBGrid.SetUpColumns;
      (* Everything else must be loaded *)

      begin (* tRRFDBGrid.SetUpColumns *)
        ParseOptions;
        SetColumnTitleStyle
      end   (* tRRFDBGrid.SetUpColumns *);

    procedure tRRFDBGrid.TitleClick(Column: tColumn);
      (* Select an index.  That is,
            0. if the DataSource is not a table (that is, is a query), then
                 quit
            1. isolate the fieldname <FN> = Column.Fieldname of the clicked column
            2. look through SortFieldNames.Strings for <FN>
            3. If the corresponding member of SortIndexNames is not the current
                  index, then make it be the current index.
      *)

      var
        TheEntryIndex: integer;

      begin (* tRRFDBGrid.TitleClick *)
        if DataSource.DataSet is tTable then
          begin (* nondegenerate *)
            TheEntryIndex := SortFieldNames.IndexOf(Column.FieldName);
            if TheEntryIndex >= 0 then
              with tTable(DataSource.DataSet) do
                if IndexName <> SortIndexNames[TheEntryIndex] then
                  IndexName := SortIndexNames[TheEntryIndex]
          end   (* nondegenerate *)
        else
          ShowMessage('RRFDBGrid cannot change sort order of query-based display.')
      end   (* tRRFDBGrid.TitleClick *);

    procedure tRRFDBGrid.ToggleBoolean;
      (* Toggle the boolean field *)

      begin (* TRRFDBGrid.ToggleBoolean *)
        if not ReadOnly then
          with SelectedField do
            begin (* nondegenerate *)
              DataSet.Edit;
              AsBoolean := not AsBoolean;
              DataSet.Post
          end   (* nondegenerate *)
      end   (* TRRFDBGrid.ToggleBoolean *);
{
end.begin(950403)
last=1868, 0=-1, 1=-1, 2=-1, 3=-1, 4=-1, 5=-1, 6=-1, 7=-1, 8=-1, 9=-1
cf(E_SCRL_MARG,0,LOCAL) cf(PG_AUTO_IND,1,LOCAL) cf(PG_IND_INC,1,LOCAL) cf(W_LF_MARG,0,LOCAL) cf(W_RT_MARG,0,LOCAL) cf(W_WORD_WRAP,0,LOCAL) cft(    8;LOCAL)
last modified 06-05-99 13:22
end.end(950403)
}

end.

